class ParclLabsError(Exception):
    """Base exception class for ParclLabs SDK"""

    pass


class NotFoundError(ParclLabsError):
    """Exception raised when no data is found matching search criteria (404 error)."""

    def __init__(
        self,
        message: str = (
            "No data found matching search criteria. Try a different set of parameters."
        ),
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(message, *args, **kwargs)


class DataValidationError(ParclLabsError):
    """Exception raised for data validation errors (422 error)."""

    def __init__(
        self,
        message: str = "Data validation error occurred.",
        details: object = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        self.details = details
        super().__init__(message, *args, **kwargs)

    def __str__(self) -> str:
        if self.details:
            return f"{super().__str__()}\nDetails: {self.details}"
        return super().__str__()



from enum import Enum
from typing import Any


class RequestMethods(Enum):
    GET = "GET"
    POST = "POST"


class RequestLimits(Enum):
    DEFAULT = 12
    MAX_POST = 1000
    DEFAULT_SMALL = 1000
    DEFAULT_LARGE = 10000
    PROPERTY_V2_MAX = 50000


class ResponseCodes(Enum):
    SUCCESS = 200
    CLIENT_ERROR = 400
    FORBIDDEN = 403
    NOT_FOUND = 404
    VALIDATION_ERROR = 422
    RATE_LIMIT_EXCEEDED = 429
    SERVER_ERROR = 500


class PropertyTypes(Enum):
    SINGLE_FAMILY = "SINGLE_FAMILY"
    CONDO = "CONDO"
    TOWNHOUSE = "TOWNHOUSE"
    OTHER = "OTHER"
    ALL_PROPERTIES = "ALL_PROPERTIES"


class PropertyTypesUnit(Enum):
    SINGLE_FAMILY = "SINGLE_FAMILY"
    CONDO = "CONDO"
    TOWNHOUSE = "TOWNHOUSE"
    OTHER = "OTHER"


class EntityNames(Enum):
    AMH = "AMH"
    TRICON = "TRICON"
    INVITATION_HOMES = "INVITATION_HOMES"
    HOME_PARTNERS_OF_AMERICA = "HOME_PARTNERS_OF_AMERICA"
    PROGRESS_RESIDENTIAL = "PROGRESS_RESIDENTIAL"
    FIRSTKEY_HOMES = "FIRSTKEY_HOMES"
    AMHERST = "AMHERST"
    VINEBROOK_HOMES = "VINEBROOK_HOMES"
    MAYMONT_HOMES = "MAYMONT_HOMES"
    SFR3 = "SFR3"


class PortfolioSizes(Enum):
    PORTFOLIO_2_TO_9 = "PORTFOLIO_2_TO_9"
    PORTFOLIO_10_TO_99 = "PORTFOLIO_10_TO_99"
    PORTFOLIO_100_TO_999 = "PORTFOLIO_100_TO_999"
    PORTFOLIO_1000_PLUS = "PORTFOLIO_1000_PLUS"
    ALL_PORTFOLIOS = "ALL_PORTFOLIOS"


class LocationTypes(Enum):
    COUNTY = "COUNTY"
    CITY = "CITY"
    ZIP5 = "ZIP5"
    CDP = "CDP"
    VILLAGE = "VILLAGE"
    TOWN = "TOWN"
    CBSA = "CBSA"
    ALL = "ALL"


class USRegions(Enum):
    EAST_NORTH_CENTRAL = "EAST_NORTH_CENTRAL"
    EAST_SOUTH_CENTRAL = "EAST_SOUTH_CENTRAL"
    MIDDLE_ATLANTIC = "MIDDLE_ATLANTIC"
    MOUNTAIN = "MOUNTAIN"
    NEW_ENGLAND = "NEW_ENGLAND"
    PACIFIC = "PACIFIC"
    SOUTH_ATLANTIC = "SOUTH_ATLANTIC"
    WEST_NORTH_CENTRAL = "WEST_NORTH_CENTRAL"
    WEST_SOUTH_CENTRAL = "WEST_SOUTH_CENTRAL"
    ALL = "ALL"


class SortByParams(Enum):
    TOTAL_POPULATION = "TOTAL_POPULATION"
    MEDIAN_INCOME = "MEDIAN_INCOME"
    CASE_SHILLER_20_MARKET = "CASE_SHILLER_20_MARKET"
    CASE_SHILLER_10_MARKET = "CASE_SHILLER_10_MARKET"
    PRICEFEED_MARKET = "PRICEFEED_MARKET"
    PARCL_EXCHANGE_MARKET = "PARCL_EXCHANGE_MARKET"


class SortOrder(Enum):
    ASC = "ASC"
    DESC = "DESC"


class EventTypes(Enum):
    SALE = "SALE"
    LISTING = "LISTING"
    RENTAL = "RENTAL"
    ALL = "ALL"


class ResponseColumns(Enum):
    PARCL_ID = "parcl_id"
    PARCL_PROPERTY_ID = "parcl_property_id"
    DATE = "date"
    EVENT_DATE = "event_date"
    EVENT_TYPE = "event_type"
    TOTAL = "total"
    LIMIT = "limit"
    OFFSET = "offset"
    LINKS = "links"
    ACCOUNT = "account"


class USStateAbbreviations(Enum):
    AK = "AK"
    AL = "AL"
    AR = "AR"
    AZ = "AZ"
    CA = "CA"
    CO = "CO"
    CT = "CT"
    DC = "DC"
    DE = "DE"
    FL = "FL"
    GA = "GA"
    HI = "HI"
    IA = "IA"
    ID = "ID"
    IL = "IL"
    IN = "IN"
    KS = "KS"
    KY = "KY"
    LA = "LA"
    MA = "MA"
    MD = "MD"
    ME = "ME"
    MI = "MI"
    MN = "MN"
    MO = "MO"
    MS = "MS"
    MT = "MT"
    NC = "NC"
    ND = "ND"
    NE = "NE"
    NH = "NH"
    NJ = "NJ"
    NM = "NM"
    NV = "NV"
    NY = "NY"
    OH = "OH"
    OK = "OK"
    OR = "OR"
    PA = "PA"
    PR = "PR"
    RI = "RI"
    SC = "SC"
    SD = "SD"
    TN = "TN"
    TX = "TX"
    UT = "UT"
    VA = "VA"
    VI = "VI"
    VT = "VT"
    WA = "WA"
    WI = "WI"
    WV = "WV"
    WY = "WY"
    ALL = "ALL"


class USStateFIPSCodes(Enum):
    ALABAMA = "01"
    ALASKA = "02"
    ARIZONA = "04"
    ARKANSAS = "05"
    CALIFORNIA = "06"
    COLORADO = "08"
    CONNECTICUT = "09"
    DELAWARE = "10"
    DISTRICT_OF_COLUMBIA = "11"
    FLORIDA = "12"
    GEORGIA = "13"
    HAWAII = "15"
    IDAHO = "16"
    ILLINOIS = "17"
    INDIANA = "18"
    IOWA = "19"
    KANSAS = "20"
    KENTUCKY = "21"
    LOUISIANA = "22"
    MAINE = "23"
    MARYLAND = "24"
    MASSACHUSETTS = "25"
    MICHIGAN = "26"
    MINNESOTA = "27"
    MISSISSIPPI = "28"
    MISSOURI = "29"
    MONTANA = "30"
    NEBRASKA = "31"
    NEVADA = "32"
    NEW_HAMPSHIRE = "33"
    NEW_JERSEY = "34"
    NEW_MEXICO = "35"
    NEW_YORK = "36"
    NORTH_CAROLINA = "37"
    NORTH_DAKOTA = "38"
    OHIO = "39"
    OKLAHOMA = "40"
    OREGON = "41"
    PENNSYLVANIA = "42"
    RHODE_ISLAND = "44"
    SOUTH_CAROLINA = "45"
    SOUTH_DAKOTA = "46"
    TENNESSEE = "47"
    TEXAS = "48"
    UTAH = "49"
    VERMONT = "50"
    VIRGINIA = "51"
    WASHINGTON = "53"
    WEST_VIRGINIA = "54"
    WISCONSIN = "55"
    WYOMING = "56"
    AMERICAN_SAMOA = "60"
    GUAM = "66"
    NORTHERN_MARIANA_ISLANDS = "69"
    PUERTO_RICO = "72"
    US_VIRGIN_ISLANDS = "78"
    ALL = "ALL"


def get_enum_values(enum_class: Enum) -> list[Any]:
    """
    Get the values of an enum class.

    Args:
        enum_class (Enum): The enum class to get the values of.

    Returns:
        List[Any]: A list of the values of the enum class.
    """

    return [item.value for item in enum_class]



import numpy as np
import pandas as pd

from parcllabs.common import DATE_COLUMNS, ID_COLUMNS


def _process_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """Process a single DataFrame to handle empty values and columns."""
    # Select only common columns
    processed_df = df.copy()

    # Remove columns that are entirely NA
    processed_df = processed_df.dropna(axis=1, how="all")

    # For columns that are all empty strings, replace with NaN
    for col in processed_df.columns:
        if (
            processed_df[col].dtype == object
            and processed_df[col].astype(str).str.strip().eq("").all()
        ):
            processed_df[col] = np.nan

    # Remove columns that became all NaN after replacing empty strings
    processed_df = processed_df.dropna(axis=1, how="all")

    return processed_df


def _reorder_columns(output: pd.DataFrame, original_columns: list) -> pd.DataFrame:
    """Reorder columns based on specified rules and original order."""
    final_columns = []

    # Rule 1: parcl_id or parcl_property_id should be the first column
    for col in ID_COLUMNS:
        if col in output.columns:
            final_columns.append(col)
            break

    # Rule 2: date or event_date should be the second column
    date_columns = ["date", "event_date"]
    for col in date_columns:
        if col in output.columns:
            final_columns.append(col)
            break

    # Add remaining columns in their original order
    for col in original_columns:
        if col in output.columns and col not in final_columns:
            final_columns.append(col)

    return output[final_columns]


def safe_concat_and_format_dtypes(data_container: list) -> pd.DataFrame:
    # Filter out empty DataFrames
    non_empty_dfs = [df for df in data_container if not df.empty]

    # Check if there are any DataFrames left to concatenate
    if not non_empty_dfs:
        return pd.DataFrame()  # Return an empty DataFrame if no non-empty DataFrames are found

    # Get the column order from the first non-empty DataFrame
    original_columns = non_empty_dfs[0].columns.tolist()

    # Find common columns across all non-empty DataFrames
    common_columns = list(set.intersection(*[set(df.columns) for df in non_empty_dfs]))

    # Process each DataFrame
    processed_dfs = []
    for df in non_empty_dfs:
        common_cols_df = df[common_columns].copy()
        processed_df = _process_dataframe(common_cols_df)
        if not processed_df.empty:
            processed_dfs.append(processed_df)

    # Concatenate the processed DataFrames
    if not processed_dfs:
        return pd.DataFrame()  # Return an empty DataFrame if all processed DataFrames are empty

    output = pd.concat(processed_dfs, axis=0, ignore_index=True)

    # Cast date columns to datetime
    for col in DATE_COLUMNS:
        if col in output.columns:
            output[col] = pd.to_datetime(output[col], errors="coerce")

    # Reorder columns
    return _reorder_columns(output, original_columns)



import json
import platform
from collections import deque
from collections.abc import Mapping
from typing import Any

import pandas as pd
import requests
from requests.exceptions import RequestException

from parcllabs.__version__ import VERSION
from parcllabs.common import DELETE_FROM_OUTPUT, GET_METHOD, POST_METHOD
from parcllabs.enums import RequestLimits, RequestMethods, ResponseCodes
from parcllabs.exceptions import NotFoundError
from parcllabs.services.data_utils import safe_concat_and_format_dtypes
from parcllabs.services.validators import Validators


class ParclLabsService:
    """
    Base class for working with data from the Parcl Labs API.
    """

    def __init__(self, url: str, client: object, post_url: str | None = None) -> None:
        self.url = url
        self.post_url = post_url
        self.client = client
        if client is None:
            raise ValueError("Missing required client object.")
        self.api_url = client.api_url
        self.full_url = self.api_url + self.url
        self.full_post_url = self.api_url + self.post_url if post_url else None
        self.api_key = client.api_key
        self.headers = self._get_headers()

    def _get_headers(self) -> dict[str, str]:
        """
        Generate the headers for API requests.

        This method constructs a dictionary of headers to be used in API requests.
        It includes authentication, content type, client version, and platform
        information.

        Returns:
            Dict[str, str]: A dictionary containing the following headers:
                - Authorization: The API key for authentication.
                - Content-Type: Set to "application/json" for JSON payloads.
                - X-Parcl-Labs-Python-Client-Version: The version of the Python client.
                - X-Parcl-Labs-Python-Client-Platform: The operating system of the
                client.
                - X-Parcl-Labs-Python-Client-Platform-Version: The Python version of the
                 client.
        """
        return {
            "Authorization": f"{self.api_key}",
            "Content-Type": "application/json",
            "X-Parcl-Labs-Python-Client-Version": VERSION,
            "X-Parcl-Labs-Python-Client-Platform": platform.system(),
            "X-Parcl-Labs-Python-Client-Platform-Version": platform.python_version(),
        }

    def _clean_params(self, params: dict[str, Any]) -> dict[str, Any]:
        """
        Remove None values from the parameters dictionary.

        This method filters out key-value pairs from the input dictionary
        where the value is None. This is useful for preparing request
        parameters, ensuring that only non-None values are sent in the request.

        Args:
            params (Dict[str, Any]): The original parameters dictionary.

        Returns:
            Dict[str, Any]: A new dictionary with all None values removed.
        """
        return {k: v for k, v in params.items() if v is not None}

    def _make_request(self, method: str, url: str, **kwargs: dict) -> requests.Response:
        """
        Generic method to make HTTP requests and handle errors.

        Args:
            method (str): The HTTP method ('GET' or 'POST').
            url (str): The URL endpoint to request.
            **kwargs: Additional arguments to pass to the request method.

        Returns:
            requests.Response: The response object.

        Raises:
            RequestException: If the request fails or an unexpected error occurs.
        """
        try:
            if method == GET_METHOD:
                params = kwargs.get("params", {})
                kwargs["params"] = params

            response = requests.request(method, url, headers=self.headers, **kwargs)  # noqa: S113
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            self.error_handling(response)
        except requests.exceptions.RequestException as err:
            raise RequestException(f"Request failed: {err!s}") from err
        except Exception as e:
            raise RequestException(f"An unexpected error occurred: {e!s}") from e
        else:
            return response

    def _post(
        self,
        url: str,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
    ) -> requests.Response:
        """
        Send a POST request to the specified URL with the given data.

        Args:
            url (str): The URL endpoint to request.
            params (dict, optional): The parameters to send in the query string.
            data (dict, optional): The data to send in the request body.

        Returns:
            requests.Response: The response object.
        """
        return self._make_request(POST_METHOD, url, params=params, json=data)

    def _get(self, url: str, params: dict[str, Any] | None = None) -> requests.Response:
        """
        Send a GET request to the specified URL with the given parameters.

        Args:
            url (str): The URL endpoint to request.
            params (dict, optional): The parameters to send in the query string.

        Returns:
            requests.Response: The response object.
        """
        return self._make_request(GET_METHOD, url, params=params)

    def _fetch(
        self,
        parcl_ids: list[int],
        params: Mapping[str, Any] | None,
        auto_paginate: bool = False,
    ) -> object:
        """
        This method handles the fetching of data based on the provided Parcl IDs and
        parameters.
        It supports both GET and POST requests, depending on the client's configuration
        and the number of Parcl IDs.
        Args:
            parcl_ids (List[int]): A list of Parcl IDs to fetch data for.
            params (Optional[Mapping[str, Any]]): Additional parameters for the request.
                If not provided or None, no additional parameters will be used.
            auto_paginate (bool, optional): Whether to automatically handle pagination.
                Defaults to False.
        Returns:
            The result of the fetch operation. The exact return type depends on the
            specific fetch method called (_fetch_post, _fetch_get, or
            _fetch_get_many_parcl_ids).
        """
        params = self._clean_params(params) if params else {}

        # Use client's default limit if no limit specified in params
        if "limit" not in params or params["limit"] is None:
            params["limit"] = self.client.limit

        if self.full_post_url:
            # convert the list of parcl_ids into post body params, formatted
            # as strings
            if params.get("limit"):
                params["limit"] = self._validate_limit(POST_METHOD, params["limit"])

            data = {"parcl_id": [str(pid) for pid in parcl_ids], **params}
            params = {"limit": params["limit"]} if params.get("limit") else {}

            print(f"data: {data}, params: {params}")

            return self._fetch_post(params, data, auto_paginate)
        if params.get("limit"):
            params["limit"] = self._validate_limit(GET_METHOD, params["limit"])

        if len(parcl_ids) == 1:
            url = self.full_url.format(parcl_id=parcl_ids[0])
            return self._fetch_get(url, params, auto_paginate)
        return self._fetch_get_many_parcl_ids(parcl_ids, params, auto_paginate)

    def _fetch_get_many_parcl_ids(
        self, parcl_ids: list[int], params: dict[str, Any], auto_paginate: bool
    ) -> list[dict[str, Any]]:
        """
        Fetch data for multiple Parcl IDs using individual GET requests.

        This method iterates through a list of Parcl IDs, making a separate GET request
        for each ID. It's used when multiple Parcl IDs need to be fetched and the API
        doesn't support bulk fetching in a single request.

        Args:
            parcl_ids (List[int]): A list of Parcl IDs to fetch data for.
            params (Dict[str, Any]): Additional parameters to include in each GET
            request.
            auto_paginate (bool): Whether to automatically handle pagination for each
            request.

        Returns:
            List[Dict[str, Any]]: A list of dictionaries, where each dictionary contains
            the data fetched for a single Parcl ID. The order of the results corresponds
            to the order of the input parcl_ids.
        """
        results = []

        for parcl_id in parcl_ids:
            try:
                url = self.full_url.format(parcl_id=parcl_id)
                result = self._fetch_get(url, params, auto_paginate)
                results.append(result)
            except NotFoundError:
                continue

        return results

    def _fetch_post(
        self,
        params: dict[str, Any],
        data: dict[str, Any],
        auto_paginate: bool,
    ) -> object:
        response = self._post(self.full_post_url, params=params, data=data)
        return self._process_and_paginate_response(
            response,
            auto_paginate,
            original_params=params,
            data=data,
            referring_method="post",
        )

    def _fetch_get(self, url: str, params: dict[str, Any], auto_paginate: bool) -> object:
        response = self._get(url, params=params)
        return self._process_and_paginate_response(
            response, auto_paginate, original_params=params, referring_method="get"
        )

    def _process_and_paginate_response(
        self,
        response: requests.Response,
        auto_paginate: bool,
        original_params: dict[str, Any],
        data: dict[str, Any] | None = None,
        referring_method: str = "get",
    ) -> object:
        if response.status_code == ResponseCodes.NOT_FOUND.value:
            return None
        response.raise_for_status()
        result = response.json()

        if auto_paginate and "links" in result and result["links"].get("next") is not None:
            all_items = result["items"]
            while result["links"].get("next") is not None:
                next_url = result["links"]["next"]
                if referring_method == "post":
                    next_response = self._post(next_url, data=data, params=original_params)
                else:
                    next_response = self._get(next_url, params=original_params)
                next_response.raise_for_status()
                result = next_response.json()
                all_items.extend(result["items"])
            result["items"] = all_items

        return result

    def retrieve(
        self,
        parcl_ids: list[int],
        start_date: str | None = None,
        end_date: str | None = None,
        limit: int | None = None,
        params: Mapping[str, Any] | None = None,
        auto_paginate: bool = False,
    ) -> pd.DataFrame:
        start_date = Validators.validate_date(start_date)
        end_date = Validators.validate_date(end_date)

        params = self._clean_params(
            {
                "start_date": start_date,
                "end_date": end_date,
                "limit": limit if limit else None,
                **(params or {}),
            }
        )

        data_container = []
        max_parcl_ids = 1000
        for i in range(0, len(parcl_ids), max_parcl_ids):
            try:
                chunk = parcl_ids[i : i + max_parcl_ids]
                results = self._fetch(chunk, params, auto_paginate=auto_paginate)
                data_container.extend(results if isinstance(results, list) else [results])
            except NotFoundError:
                # we don't want to kill the entire process if one of the chunks fails
                # due to no data. sparse parcl_ids can result in no data found.
                # The post request can handle all 10k in one request, however the get
                # request is one by one. get handles this direclty per parcl_id, while
                # post handles all at once.
                continue

        return self._as_pd_dataframe(data_container)

    def _update_account_info(self, account_info: dict) -> None:
        """
        Update the account info for the client.
        """
        if account_info:
            final_account_dict = {
                "est_credits_used": account_info.get("est_credits_used", 0),
                "est_session_credits_used": (
                    self.client.account_info.get("est_session_credits_used", 0)
                    + account_info.get("est_credits_used", 0)
                ),
                "est_remaining_credits": account_info.get("est_remaining_credits", 0),
            }
            self.client.account_info = final_account_dict

    @staticmethod
    def sanitize_output(data: dict[str, Any]) -> dict[str, Any]:
        return {k: v for k, v in data.items() if k not in DELETE_FROM_OUTPUT}

    def _as_pd_dataframe(self, data: list[Mapping[str, Any]]) -> pd.DataFrame:
        data_container = deque()
        for results in data:
            if results is None:
                continue
            account_info = results.get("account")
            sanitized_results = self.sanitize_output(results)
            meta_fields = [k for k in sanitized_results.keys() if k != "items"]
            normalized_df = pd.json_normalize(
                sanitized_results, record_path="items", meta=meta_fields
            )
            updated_cols_names = [c.replace(".", "_") for c in normalized_df.columns.tolist()]
            normalized_df.columns = updated_cols_names
            data_container.append(normalized_df)
            self._update_account_info(account_info)

        return safe_concat_and_format_dtypes(data_container)

    @staticmethod
    def error_handling(response: requests.Response) -> None:
        try:
            error_details = response.json()
            error_message = error_details.get("detail", "No detail provided by API")
            if response.status_code == ResponseCodes.FORBIDDEN.value:
                error_message += (
                    " Visit https://dashboard.parcllabs.com for more information"
                    " or reach out to team@parcllabs.com."
                )
            elif response.status_code == ResponseCodes.VALIDATION_ERROR.value:
                details = error_details.get("detail")
                error_message = details[0].get("msg", "validation error")
            elif response.status_code == ResponseCodes.RATE_LIMIT_EXCEEDED.value:
                error_message = error_details.get("error", "Rate Limit Exceeded")
            elif response.status_code == ResponseCodes.NOT_FOUND.value:
                raise NotFoundError(
                    "No data found matching search criteria.Try a different set of parameters."
                )

        except json.JSONDecodeError:
            error_message = "Failed to decode JSON error response"

        type_of_error = (
            "Client"
            if ResponseCodes.CLIENT_ERROR.value
            <= response.status_code
            < ResponseCodes.SERVER_ERROR.value
            else "Server"
        )
        msg = f"{response.status_code} {type_of_error} Error: {error_message}"
        raise requests.RequestException(msg)

    @staticmethod
    def _validate_limit(method: RequestMethods, limit: int) -> int:
        default_large = RequestLimits.DEFAULT_LARGE.value
        default_small = RequestLimits.DEFAULT_SMALL.value

        if method == POST_METHOD:
            if limit > default_large:
                print(
                    f"Supplied limit value is too large for requested endpoint."
                    f"Setting limit to maxium value of {default_large}."
                )
                limit = default_large

        elif method == GET_METHOD:
            if limit > default_small:
                print(
                    f"Supplied limit value is too large for requested endpoint."
                    f"Setting limit to maxium value of {default_small}."
                )
                limit = default_small

        else:
            raise ValueError("Invalid method. Must be either 'GET' or 'POST'.")

        return limit



from parcllabs.enums import (
    RequestMethods,
    ResponseColumns,
    USStateAbbreviations,
    get_enum_values,
)

VALID_US_STATE_ABBREV = get_enum_values(USStateAbbreviations)

ID_COLUMNS = [ResponseColumns.PARCL_ID.value, ResponseColumns.PARCL_PROPERTY_ID.value]
DATE_COLUMNS = [ResponseColumns.DATE.value, ResponseColumns.EVENT_DATE.value]

DELETE_FROM_OUTPUT = [
    ResponseColumns.TOTAL.value,
    ResponseColumns.LIMIT.value,
    ResponseColumns.OFFSET.value,
    ResponseColumns.LINKS.value,
    ResponseColumns.ACCOUNT.value,
]

PARCL_LABS_DASHBOARD_URL = "https://dashboard.parcllabs.com/signup"
NO_API_KEY_ERROR = (
    f"API Key is required. Please visit {PARCL_LABS_DASHBOARD_URL} to get an API key."
)

GET_METHOD = RequestMethods.GET.value
POST_METHOD = RequestMethods.POST.value


ZIP_CODE_LENGTH = 5
PARCL_PROPERTY_IDS_LIMIT = 10000
PARCL_PROPERTY_IDS = "parcl_property_ids"



import time
from collections.abc import Mapping
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any

import pandas as pd

from parcllabs.common import PARCL_PROPERTY_IDS, PARCL_PROPERTY_IDS_LIMIT
from parcllabs.enums import RequestLimits
from parcllabs.schemas.schemas import PropertyV2RetrieveParamCategories, PropertyV2RetrieveParams
from parcllabs.services.parcllabs_service import ParclLabsService
from parcllabs.services.validators import Validators


class PropertyV2Service(ParclLabsService):
    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)
        self.simple_bool_validator = Validators.validate_input_bool_param_simple

    @staticmethod
    def _raise_http_error(chunk_num: int, status_code: int, response_preview: str) -> None:
        error_msg = f"Chunk {chunk_num} failed: HTTP {status_code}"
        raise RuntimeError(f"{error_msg}\nResponse content: {response_preview}...")

    @staticmethod
    def _raise_empty_response_error(chunk_num: int) -> None:
        raise RuntimeError(f"Chunk {chunk_num} failed: Empty response from API")

    def _fetch_post(self, params: dict[str, Any], data: dict[str, Any]) -> list[dict]:
        """Fetch data using POST request with pagination support."""
        response = self._post(url=self.full_post_url, data=data, params=params)
        result = response.json()

        all_data = [result]
        pagination = result.get("pagination")
        metadata = result.get("metadata")

        if pagination:
            limit = pagination.get("limit")
            returned_count = metadata.get("results", {}).get("returned_count", 0)
            # if we got fewer or equal results than requested, don't paginate
            if returned_count <= limit:
                return all_data

            # If we need to paginate, use concurrent requests
            if pagination.get("has_more"):
                print("More pages to fetch, paginating additional pages...")
                offset = pagination.get("offset")
                total_available = metadata.get("results", {}).get("total_available", 0)

                # Calculate how many more pages we need
                remaining_pages = (total_available - limit) // limit
                if (total_available - limit) % limit > 0:
                    remaining_pages += 1

                # Generate all the URLs we need to fetch
                urls = []
                current_offset = offset + limit
                for _ in range(remaining_pages):
                    urls.append(f"{self.full_post_url}?limit={limit}&offset={current_offset}")
                    current_offset += limit

                # Use ThreadPoolExecutor to make concurrent requests
                with ThreadPoolExecutor(max_workers=self.client.num_workers) as executor:
                    future_to_url = {
                        executor.submit(self._post, url=url, data=data, params=params): url
                        for url in urls
                    }

                    for future in as_completed(future_to_url):
                        try:
                            response = future.result()
                            page_result = response.json()
                            all_data.append(page_result)
                        except Exception as exc:
                            print(f"Request failed: {exc}")

        return all_data

    def _fetch_post_parcl_property_ids(
        self,
        params: dict[str, Any],
        data: dict[str, Any],
    ) -> list[dict]:
        """Fetch data using POST request with parcl_property_ids, chunking the request

        Args:
            params: Dictionary of parameters to pass to the request.
            data: Dictionary of data to pass to the request.

        Returns:
            List of dictionaries containing the data from the request.
        """
        parcl_property_ids = data.get(PARCL_PROPERTY_IDS)
        num_ids = len(parcl_property_ids)
        if num_ids <= PARCL_PROPERTY_IDS_LIMIT:
            return self._fetch_post(params=params, data=data)

        # If we exceed PARCL_PROPERTY_IDS_LIMIT, chunk the request
        parcl_property_ids_chunks = [
            parcl_property_ids[i : i + PARCL_PROPERTY_IDS_LIMIT]
            for i in range(0, num_ids, PARCL_PROPERTY_IDS_LIMIT)
        ]
        num_chunks = len(parcl_property_ids_chunks)

        print(f"Fetching {num_chunks} chunks...")

        all_data = []
        with ThreadPoolExecutor(max_workers=3) as executor:
            # Create a copy of data for each chunk to avoid race conditions
            future_to_chunk = {}
            for idx, chunk in enumerate(parcl_property_ids_chunks):
                # Create a copy of data with the specific chunk
                chunk_data = data.copy()
                chunk_data[PARCL_PROPERTY_IDS] = chunk

                # Submit the task
                future = executor.submit(
                    self._post,
                    url=self.full_post_url,
                    data=chunk_data,
                    params=params,
                )
                future_to_chunk[future] = idx + 1

                # Small delay between submissions to avoid rate limiting
                if idx < len(parcl_property_ids_chunks) - 1:  # Don't delay after the last one
                    time.sleep(0.1)

            # Helper functions to abstract raise statements

            # Collect results as they complete
            for future in as_completed(future_to_chunk):
                chunk_num = future_to_chunk[future]
                try:
                    result = future.result()

                    # Check HTTP status code
                    if result.status_code != 200:
                        response_preview = (
                            result.text[:200] if result.text else "No response content"
                        )
                        self._raise_http_error(chunk_num, result.status_code, response_preview)

                    # Check if response has content
                    if not result.text.strip():
                        self._raise_empty_response_error(chunk_num)

                    # Try to parse JSON
                    try:
                        response = result.json()
                        all_data.append(response)
                        print(f"Completed chunk {chunk_num} of {num_chunks}")
                    except ValueError as json_exc:
                        response_preview = (
                            result.text[:200] if result.text else "No response content"
                        )
                        raise RuntimeError(
                            f"Chunk {chunk_num} failed: Invalid JSON - {json_exc}\n"
                            f"Response content: {response_preview}..."
                        ) from json_exc

                except Exception as exc:
                    # If it's already a RuntimeError from above, re-raise it
                    if isinstance(exc, RuntimeError):
                        raise

                    # For any other unexpected errors, wrap and raise
                    raise RuntimeError(
                        f"Chunk {chunk_num} failed with unexpected error: {exc} "
                        f"(Exception type: {type(exc).__name__})"
                    ) from exc

        print(f"All {num_chunks} chunks completed successfully.")
        return all_data

    def _as_pd_dataframe(self, data: list[Mapping[str, Any]]) -> pd.DataFrame:
        """
        Convert API response data to a pandas DataFrame with events as rows
        using json_normalize.
        """
        # First, extract all properties with their events
        properties_with_events = []

        for results in data:
            if results is None or not results.get("data"):
                continue

            account_info = results.get("account_info")
            query_data = results.get("data")

            # Process each property
            for property_data in query_data:
                events = property_data.get("events", [])

                # Create a property record without events
                property_record = {k: v for k, v in property_data.items() if k != "events"}

                if not events:
                    # If no events, add the property as is
                    properties_with_events.append(property_record)
                else:
                    # For each event, create a record with property data and this event
                    for event in events:
                        combined_record = property_record.copy()
                        combined_record["event"] = event
                        properties_with_events.append(combined_record)

            self._update_account_info(account_info)

        if not properties_with_events:
            return pd.DataFrame()

        # Use json_normalize to flatten the nested structure
        all_data_df = pd.json_normalize(
            properties_with_events,
            sep="_",  # Use underscore as separator for nested fields
        )

        # If we have event data, normalize it
        if "event" in all_data_df.columns:
            # Get indices of rows with events
            event_indices = all_data_df["event"].notna()

            if event_indices.any():
                # Normalize the event data
                event_df = pd.json_normalize(
                    all_data_df.loc[event_indices, "event"].tolist(), sep="_"
                )

                # Add event_ prefix to all columns
                event_df.columns = ["event_" + col for col in event_df.columns]

                # Add the event data back to the main dataframe
                for col in event_df.columns:
                    all_data_df.loc[event_indices, col] = event_df[col].to_numpy()

                # Drop the original event column
                all_data_df = all_data_df.drop("event", axis=1)

        return all_data_df

    def _get_metadata(self, results: list[Mapping[str, Any]]) -> dict[str, Any]:
        """Get metadata from results with accurate returned_count."""
        if not results:
            return {}

        # Start with a copy of the first result's metadata
        metadata = results[0].get("metadata", {}).copy()

        # Calculate total returned_count
        total_returned = sum(
            result.get("metadata", {}).get("results", {}).get("returned_count", 0)
            for result in results
        )
        if "results" in metadata:
            metadata["results"]["returned_count"] = total_returned

        return metadata

    def _build_search_criteria(
        self,
        parcl_ids: list[int] | None = None,
        parcl_property_ids: list[int] | None = None,
        geo_coordinates: dict[str, float] | None = None,
    ) -> dict[str, Any]:
        """Build and validate search criteria."""
        data = {}

        if parcl_ids:
            data["parcl_ids"] = parcl_ids

        if parcl_property_ids:
            data[PARCL_PROPERTY_IDS] = parcl_property_ids

        if geo_coordinates:
            data["geo_coordinates"] = geo_coordinates

        return data

    def _build_numeric_filters(self, params: PropertyV2RetrieveParams) -> dict[str, Any]:
        """Build numeric property filters."""
        filters = {}

        if params.min_beds is not None:
            filters["min_beds"] = params.min_beds
        if params.max_beds is not None:
            filters["max_beds"] = params.max_beds
        if params.min_baths is not None:
            filters["min_baths"] = params.min_baths
        if params.max_baths is not None:
            filters["max_baths"] = params.max_baths
        if params.min_sqft is not None:
            filters["min_sqft"] = params.min_sqft
        if params.max_sqft is not None:
            filters["max_sqft"] = params.max_sqft
        if params.min_year_built is not None:
            filters["min_year_built"] = params.min_year_built
        if params.max_year_built is not None:
            filters["max_year_built"] = params.max_year_built

        return filters

    def _build_date_filters(self, params: PropertyV2RetrieveParams) -> dict[str, Any]:
        """Build date-related property filters."""
        filters = {}

        if params.min_record_added_date is not None:
            filters["min_record_added_date"] = params.min_record_added_date
        if params.max_record_added_date is not None:
            filters["max_record_added_date"] = params.max_record_added_date

        return filters

    def _build_boolean_filters(self, params: PropertyV2RetrieveParams) -> dict[str, Any]:
        """Build boolean property filters."""
        filters = {}

        if params.include_property_details is not None:
            filters["include_property_details"] = self.simple_bool_validator(
                params.include_property_details
            )
        if params.current_on_market_flag is not None:
            filters["current_on_market_flag"] = self.simple_bool_validator(
                params.current_on_market_flag
            )
        if params.current_on_market_rental_flag is not None:
            filters["current_on_market_rental_flag"] = self.simple_bool_validator(
                params.current_on_market_rental_flag
            )
        if params.current_new_construction_flag is not None:
            filters["current_new_construction_flag"] = self.simple_bool_validator(
                params.current_new_construction_flag
            )
        if params.current_owner_occupied_flag is not None:
            filters["current_owner_occupied_flag"] = self.simple_bool_validator(
                params.current_owner_occupied_flag
            )
        if params.current_investor_owned_flag is not None:
            filters["current_investor_owned_flag"] = self.simple_bool_validator(
                params.current_investor_owned_flag
            )

        return filters

    def _build_property_filters(self, params: PropertyV2RetrieveParams) -> dict[str, Any]:
        """Build property filters from validated Pydantic schema."""
        property_filters = {}

        # Build numeric filters
        property_filters.update(self._build_numeric_filters(params))

        # Build date filters
        property_filters.update(self._build_date_filters(params))

        # Build boolean filters
        property_filters.update(self._build_boolean_filters(params))

        # Handle property types
        if params.property_types:
            property_filters["property_types"] = [
                property_type.upper() for property_type in params.property_types
            ]

        # Handle current entity owner name
        if params.current_entity_owner_name is not None:
            property_filters["current_entity_owner_name"] = params.current_entity_owner_name

        return property_filters

    def _build_event_filters(self, params: PropertyV2RetrieveParams) -> dict[str, Any]:  # noqa: C901
        """Build event filters from validated Pydantic schema."""
        event_filters = {}

        # Handle event names
        if params.event_names:
            event_filters["event_names"] = [event_name.upper() for event_name in params.event_names]

        # Handle date and price filters
        if params.min_event_date is not None:
            event_filters["min_event_date"] = params.min_event_date
        if params.max_event_date is not None:
            event_filters["max_event_date"] = params.max_event_date
        if params.min_record_updated_date is not None:
            event_filters["min_record_updated_date"] = params.min_record_updated_date
        if params.max_record_updated_date is not None:
            event_filters["max_record_updated_date"] = params.max_record_updated_date
        if params.min_price is not None:
            event_filters["min_price"] = params.min_price
        if params.max_price is not None:
            event_filters["max_price"] = params.max_price

        # Handle boolean parameters
        if params.is_new_construction is not None:
            event_filters["is_new_construction"] = self.simple_bool_validator(
                params.is_new_construction
            )
        if params.include_events is not None:
            event_filters["include_events"] = self.simple_bool_validator(params.include_events)
        if params.include_full_event_history is not None:
            event_filters["include_full_event_history"] = self.simple_bool_validator(
                params.include_full_event_history
            )

        return event_filters

    def _build_owner_filters(self, params: PropertyV2RetrieveParams) -> dict[str, Any]:
        """Build owner filters from validated Pydantic schema."""
        owner_filters = {}

        # Handle owner names
        if params.owner_name:
            owner_filters["owner_name"] = [owner_name.upper() for owner_name in params.owner_name]

        # Handle entity seller names
        if params.entity_seller_name:
            owner_filters["entity_seller_name"] = [
                entity_seller_name.upper() for entity_seller_name in params.entity_seller_name
            ]

        # Handle boolean parameters
        if params.is_current_owner is not None:
            owner_filters["is_current_owner"] = self.simple_bool_validator(params.is_current_owner)
        if params.is_investor_owned is not None:
            owner_filters["is_investor_owned"] = self.simple_bool_validator(
                params.is_investor_owned
            )
        if params.is_owner_occupied is not None:
            owner_filters["is_owner_occupied"] = self.simple_bool_validator(
                params.is_owner_occupied
            )

        return owner_filters

    def _validate_limit(self, limit: int | None) -> int:
        """Validate limit parameter."""
        max_limit = RequestLimits.PROPERTY_V2_MAX.value

        # If auto-paginate is enabled or no limit is provided, use maximum limit
        if limit in (None, 0):
            print(f"No limit provided. Setting limit to maximum value of {max_limit}.")
            return max_limit

        # If limit exceeds maximum, cap it
        if limit > max_limit:
            print(
                f"Supplied limit value is too large for requested endpoint."
                f"Setting limit to maximum value of {max_limit}."
            )
            return max_limit

        return limit

    def _build_param_categories(
        self, params: PropertyV2RetrieveParams
    ) -> PropertyV2RetrieveParamCategories:
        """Build parameter categories from validated Pydantic schema."""
        return PropertyV2RetrieveParamCategories(
            property_filters=self._build_property_filters(params),
            event_filters=self._build_event_filters(params),
            owner_filters=self._build_owner_filters(params),
        )

    def retrieve(
        self,
        parcl_ids: list[int] | None = None,
        parcl_property_ids: list[int] | None = None,
        geo_coordinates: dict[str, float] | None = None,
        property_types: list[str] | None = None,
        min_beds: int | None = None,
        max_beds: int | None = None,
        min_baths: float | None = None,
        max_baths: float | None = None,
        min_sqft: int | None = None,
        max_sqft: int | None = None,
        min_year_built: int | None = None,
        max_year_built: int | None = None,
        include_property_details: bool | None = None,
        min_record_added_date: str | None = None,
        max_record_added_date: str | None = None,
        event_names: list[str] | None = None,
        min_event_date: str | None = None,
        max_event_date: str | None = None,
        min_price: int | None = None,
        max_price: int | None = None,
        is_new_construction: bool | None = None,
        min_record_updated_date: str | None = None,
        max_record_updated_date: str | None = None,
        is_current_owner: bool | None = None,
        owner_name: list[str] | None = None,
        entity_seller_name: list[str] | None = None,
        is_investor_owned: bool | None = None,
        is_owner_occupied: bool | None = None,
        current_on_market_flag: bool | None = None,
        current_on_market_rental_flag: bool | None = None,
        current_new_construction_flag: bool | None = None,
        current_owner_occupied_flag: bool | None = None,
        current_investor_owned_flag: bool | None = None,
        current_entity_owner_name: str | None = None,
        include_events: bool | None = None,
        include_full_event_history: bool | None = None,
        limit: int | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> tuple[pd.DataFrame, dict[str, Any]]:
        """
        Retrieve property data based on search criteria and filters.

        Args:
            parcl_ids: List of parcl_ids to filter by.
            parcl_property_ids: List of parcl_property_ids to filter by.
            geo_coordinates: Dictionary containing latitude, longitude, and radius (in miles)
            to filter by.
            property_types: List of property types to filter by.
            min_beds: Minimum number of bedrooms to filter by.
            max_beds: Maximum number of bedrooms to filter by.
            min_baths: Minimum number of bathrooms to filter by.
            max_baths: Maximum number of bathrooms to filter by.
            min_sqft: Minimum square footage to filter by.
            max_sqft: Maximum square footage to filter by.
            min_year_built: Minimum year built to filter by.
            max_year_built: Maximum year built to filter by.
            include_property_details: Whether to include property details.
            min_record_added_date: Minimum record added date to filter by.
            max_record_added_date: Maximum record added date to filter by.
            event_names: List of event names to filter by.
            min_event_date: Minimum event date to filter by.
            max_event_date: Maximum event date to filter by.
            min_price: Minimum price to filter by.
            max_price: Maximum price to filter by.
            is_new_construction: Whether to filter by new construction.
            min_record_updated_date: Minimum record updated date to filter by.
            max_record_updated_date: Maximum record updated date to filter by.
            is_current_owner: Whether to filter by current owner.
            owner_name: List of owner names to filter by.
            entity_seller_name: List of entity seller names to filter by.
            is_investor_owned: Whether to filter by investor owned.
            is_owner_occupied: Whether to filter by owner occupied.
            current_on_market_flag: Whether to filter by current_on_market flag.
            current_on_market_rental_flag: Whether to filter by current_on_market_rental flag.
            current_new_construction_flag: Whether to filter by current_new_construction flag.
            current_owner_occupied_flag: Whether to filter by current_owner_occupied flag.
            current_investor_owned_flag: Whether to filter by current_investor_owned flag.
            current_entity_owner_name: Current entity owner name to filter by.
            include_events: Whether to include events in the response.
            include_full_event_history: Whether to include full event history in the response.
            limit: Number of results to return.
            params: Additional parameters to pass to the request.
        Returns:
            A tuple containing (pandas DataFrame, metadata dictionary).
        """
        print("Processing property search request...")

        # Validate and process input parameters using Pydantic schema
        input_params = PropertyV2RetrieveParams(
            parcl_ids=parcl_ids,
            parcl_property_ids=parcl_property_ids,
            geo_coordinates=geo_coordinates,
            property_types=property_types,
            min_beds=min_beds,
            max_beds=max_beds,
            min_baths=min_baths,
            max_baths=max_baths,
            min_sqft=min_sqft,
            max_sqft=max_sqft,
            min_year_built=min_year_built,
            max_year_built=max_year_built,
            include_property_details=include_property_details,
            min_record_added_date=min_record_added_date,
            max_record_added_date=max_record_added_date,
            event_names=event_names,
            min_event_date=min_event_date,
            max_event_date=max_event_date,
            min_price=min_price,
            max_price=max_price,
            is_new_construction=is_new_construction,
            min_record_updated_date=min_record_updated_date,
            max_record_updated_date=max_record_updated_date,
            is_current_owner=is_current_owner,
            owner_name=owner_name,
            entity_seller_name=entity_seller_name,
            is_investor_owned=is_investor_owned,
            is_owner_occupied=is_owner_occupied,
            current_on_market_flag=current_on_market_flag,
            current_on_market_rental_flag=current_on_market_rental_flag,
            current_new_construction_flag=current_new_construction_flag,
            current_owner_occupied_flag=current_owner_occupied_flag,
            current_investor_owned_flag=current_investor_owned_flag,
            current_entity_owner_name=current_entity_owner_name,
            include_events=include_events,
            include_full_event_history=include_full_event_history,
            limit=limit,
            params=params or {},
        )

        # Build search criteria
        data = self._build_search_criteria(
            parcl_ids=input_params.parcl_ids,
            parcl_property_ids=input_params.parcl_property_ids,
            geo_coordinates=(
                input_params.geo_coordinates.model_dump() if input_params.geo_coordinates else None
            ),
        )

        # Build parameter categories using validated parameters
        param_categories = self._build_param_categories(input_params)

        # Update data with categories
        data.update(param_categories.model_dump(exclude_none=True))

        # Set limit
        request_params = input_params.params.copy()

        # Make request with params
        if data.get(PARCL_PROPERTY_IDS):
            request_params["limit"] = PARCL_PROPERTY_IDS_LIMIT
            results = self._fetch_post_parcl_property_ids(params=request_params, data=data)
        else:
            request_params["limit"] = self._validate_limit(input_params.limit)
            results = self._fetch_post(params=request_params, data=data)

        # Get metadata from results
        metadata = self._get_metadata(results)

        # Process results
        final_df = self._as_pd_dataframe(results)

        return final_df, metadata



from collections.abc import Mapping
from typing import Any

import pandas as pd

from parcllabs.common import (
    GET_METHOD,
)
from parcllabs.services.parcllabs_service import ParclLabsService


class SearchMarkets(ParclLabsService):
    """
    Retrieve parcl_id and metadata for geographic markets in the Parcl Labs API.
    """

    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)

    def _as_pd_dataframe(self, data: list[Mapping[str, Any]]) -> pd.DataFrame:
        final_df = pd.DataFrame(data)
        # Convert numpy integers to regular Python integers
        for col in final_df.columns:
            if final_df[col].dtype in ["int64", "int32", "int16", "int8"]:
                final_df[col] = final_df[col].astype(int)
        return final_df

    def _prepare_params(
        self,
        query: str | None = None,
        location_type: str | None = None,
        region: str | None = None,
        state_abbreviation: str | None = None,
        state_fips_code: str | None = None,
        parcl_id: int | None = None,
        geoid: str | None = None,
        sort_by: str | None = None,
        sort_order: str | None = None,
        limit: int | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> dict[str, Any]:
        params = params or {}

        # Handle uppercase parameters
        uppercase_params = {
            "location_type": location_type,
            "region": region,
            "state_abbreviation": state_abbreviation,
            "state_fips_code": state_fips_code,
            "sort_by": sort_by,
            "sort_order": sort_order,
        }

        for key, value in uppercase_params.items():
            if value:
                params[key] = value.upper()

        # Handle regular parameters
        if query:
            params["query"] = query
        if parcl_id:
            params["parcl_id"] = parcl_id
        if geoid:
            params["geoid"] = geoid

        # Handle limit parameter
        if limit:
            params["limit"] = self._validate_limit(GET_METHOD, limit)
        elif self.client.limit:
            params["limit"] = self._validate_limit(GET_METHOD, self.client.limit)

        return params

    def retrieve(
        self,
        query: str | None = None,
        location_type: str | None = None,
        region: str | None = None,
        state_abbreviation: str | None = None,
        state_fips_code: str | None = None,
        parcl_id: int | None = None,
        geoid: str | None = None,
        sort_by: str | None = None,
        sort_order: str | None = None,
        limit: int | None = None,
        params: Mapping[str, Any] | None = None,
        auto_paginate: bool = False,
    ) -> pd.DataFrame:
        """
        Retrieve parcl_id and metadata for geographic markets in the Parcl Labs API.

        Args:
            query (str, optional): The search query to filter results by.
            location_type (str, optional): The location type to filter results by.
            region (str, optional): The region to filter results by.
            state_abbreviation (str, optional): The state abbreviation to filter results
            by.
            state_fips_code (str, optional): The state FIPS code to filter results by.
            parcl_id (int, optional): The parcl_id to filter results by.
            geoid (str, optional): The geoid to filter results by.
            sort_by (str, optional): The field to sort results by.
            sort_order (str, optional): The sort order to apply to the results.
            limit (int, optional): The number of items to return per page.
            params (dict, optional): Additional parameters to include in the request.
            auto_paginate (bool, optional): Automatically paginate through the results.

        Returns:
            Any: The JSON response as a pandas DataFrame.
        """
        params = self._prepare_params(
            query,
            location_type,
            region,
            state_abbreviation,
            state_fips_code,
            parcl_id,
            geoid,
            sort_by,
            sort_order,
            limit,
            params,
        )

        results = self._fetch_get(url=self.full_url, params=params, auto_paginate=auto_paginate)
        data = self._as_pd_dataframe(results.get("items"))
        self._update_account_info(results.get("account"))
        self.markets = data
        return data



from collections.abc import Mapping
from typing import Any

import pandas as pd

from parcllabs.services.parcllabs_service import ParclLabsService


class PortfolioSizeService(ParclLabsService):
    def retrieve(
        self,
        parcl_ids: list[int],
        start_date: str | None = None,
        end_date: str | None = None,
        portfolio_size: str | None = None,
        limit: int | None = None,
        params: Mapping[str, Any] | None = {},
        auto_paginate: bool = False,
    ) -> pd.DataFrame:
        """
        Retrieve portfolio size metrics for given parameters.
        """
        if portfolio_size:
            params["portfolio_size"] = portfolio_size.upper()

        return super().retrieve(
            parcl_ids=parcl_ids,
            start_date=start_date,
            end_date=end_date,
            limit=limit,
            params=params,
            auto_paginate=auto_paginate,
        )



from collections import deque
from collections.abc import Mapping
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any

import pandas as pd

from parcllabs.enums import RequestLimits
from parcllabs.exceptions import (
    NotFoundError,
)
from parcllabs.services.data_utils import (
    safe_concat_and_format_dtypes,
)
from parcllabs.services.parcllabs_service import ParclLabsService
from parcllabs.services.validators import Validators


class PropertyEventsService(ParclLabsService):
    """
    Retrieve parcl_property_id event history.
    """

    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)

    def _prepare_params(
        self,
        event_type: str | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        entity_owner_name: str | None = None,
        record_updated_date_start: str | None = None,
        record_updated_date_end: str | None = None,
    ) -> dict:
        """Prepare parameters for property events retrieval."""
        params = {}

        if event_type:
            params["event_type"] = event_type.upper()

        if entity_owner_name:
            params["entity_owner_name"] = entity_owner_name.upper()

        if start_date:
            params["start_date"] = Validators.validate_date(start_date)

        if end_date:
            params["end_date"] = Validators.validate_date(end_date)

        if record_updated_date_start:
            record_updated_date_start = Validators.validate_date(record_updated_date_start)
            params["record_updated_date_start"] = record_updated_date_start

        if record_updated_date_end:
            record_updated_date_end = Validators.validate_date(record_updated_date_end)
            params["record_updated_date_end"] = record_updated_date_end

        return params

    def retrieve(
        self,
        parcl_property_ids: list[int],
        event_type: str | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        entity_owner_name: str | None = None,
        record_updated_date_start: str | None = None,
        record_updated_date_end: str | None = None,
        params: Mapping[str, Any] | None = {},
    ) -> pd.DataFrame:
        """
        Retrieve property events for given parameters.
        """
        params = self._prepare_params(
            event_type=event_type,
            start_date=start_date,
            end_date=end_date,
            entity_owner_name=entity_owner_name,
            record_updated_date_start=record_updated_date_start,
            record_updated_date_end=record_updated_date_end,
        )

        parcl_property_ids = [str(i) for i in parcl_property_ids]
        total_properties = len(parcl_property_ids)

        def process_batch(batch_ids: list[str]) -> list[dict] | None:
            local_params = params.copy()
            local_params["parcl_property_id"] = batch_ids
            try:
                response = self._post(url=self.full_post_url, data=local_params)
                data = response.json()
                self._update_account_info(data.get("account"))
                return data.get("items")  # Return data as json
            except NotFoundError:
                return None
            except Exception as e:
                print(f"Error processing batch {batch_ids}: {e!s}")
                return None

        all_data = deque()
        with ThreadPoolExecutor(max_workers=self.client.num_workers) as executor:
            max_post_limit = RequestLimits.MAX_POST.value
            futures = {
                executor.submit(process_batch, parcl_property_ids[i : i + max_post_limit]): len(
                    parcl_property_ids[i : i + max_post_limit]
                )
                for i in range(0, total_properties, max_post_limit)
            }

            for future in as_completed(futures):
                batch_result = future.result()
                if batch_result:
                    batch_df = pd.DataFrame(batch_result)
                    all_data.append(batch_df)

        return safe_concat_and_format_dtypes(all_data)



"""
Pydantic schemas for PropertyV2Service input parameters.
"""

from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field, ValidationInfo, field_validator

from parcllabs.enums import PropertyTypes, RequestLimits


class GeoCoordinates(BaseModel):
    """Schema for geographic coordinates with radius."""

    latitude: float = Field(..., ge=-90, le=90, description="Latitude in decimal degrees")
    longitude: float = Field(..., ge=-180, le=180, description="Longitude in decimal degrees")
    radius: float = Field(..., gt=0, description="Radius in miles")


class PropertyV2RetrieveParams(BaseModel):
    """
    Input parameters schema for PropertyV2Service.retrieve() method.

    This schema validates and manages all input parameters for property searches,
    including search criteria, property filters, event filters, and owner filters.
    """

    # Search criteria
    parcl_ids: list[int] | None = Field(default=None, description="List of parcl_ids to filter by")
    parcl_property_ids: list[int] | None = Field(
        default=None, description="List of parcl_property_ids to filter by"
    )
    geo_coordinates: GeoCoordinates | None = Field(
        default=None, description="Geographic coordinates with radius to filter by"
    )

    property_types: list[str] | None = Field(
        default=None, description="List of property types to filter by"
    )
    min_beds: int | None = Field(default=None, ge=0, description="Minimum number of bedrooms")
    max_beds: int | None = Field(default=None, ge=0, description="Maximum number of bedrooms")
    min_baths: float | None = Field(default=None, ge=0, description="Minimum number of bathrooms")
    max_baths: float | None = Field(default=None, ge=0, description="Maximum number of bathrooms")
    min_sqft: int | None = Field(default=None, ge=0, description="Minimum square footage")
    max_sqft: int | None = Field(default=None, ge=0, description="Maximum square footage")
    min_year_built: int | None = Field(
        default=None, ge=1800, le=2100, description="Minimum year built"
    )
    max_year_built: int | None = Field(
        default=None, ge=1800, le=2100, description="Maximum year built"
    )
    include_property_details: bool | None = Field(
        default=None, description="Whether to include property details"
    )
    min_record_added_date: str | None = Field(
        default=None, description="Minimum record added date (YYYY-MM-DD)"
    )
    max_record_added_date: str | None = Field(
        default=None, description="Maximum record added date (YYYY-MM-DD)"
    )

    # Event filters
    include_events: bool | None = Field(
        default=None,
        description="""Boolean flag indicating if event data is included in the response.
            0 signifies that no event data is returned.
            1 signifies that event data is included, scoped
            to events that match the event filters supplied in the request
        """,
    )
    include_full_event_history: bool | None = Field(
        default=None,
        description="""Boolean flag indicating if the full event history is returned for
            each property (effective only when include_events is 1).
            0 signifies that only events matching the event filters are returned.
            1 signifies that the property's entire event history is returned
        """,
    )
    event_names: list[str] | None = Field(
        default=None, description="List of event names to filter by"
    )
    min_event_date: str | None = Field(default=None, description="Minimum event date (YYYY-MM-DD)")
    max_event_date: str | None = Field(default=None, description="Maximum event date (YYYY-MM-DD)")
    min_price: int | None = Field(default=None, ge=0, description="Minimum price")
    max_price: int | None = Field(default=None, ge=0, description="Maximum price")
    is_new_construction: bool | None = Field(
        default=None, description="Whether to filter by new construction"
    )
    min_record_updated_date: str | None = Field(
        default=None, description="Minimum record updated date (YYYY-MM-DD)"
    )
    max_record_updated_date: str | None = Field(
        default=None, description="Maximum record updated date (YYYY-MM-DD)"
    )

    # Owner filters
    is_current_owner: bool | None = Field(
        default=None, description="Whether to filter by current owner"
    )
    owner_name: list[str] | None = Field(
        default=None, description="List of owner names to filter by"
    )
    entity_seller_name: list[str] | None = Field(
        default=None, description="List of entity seller names to filter by"
    )
    is_investor_owned: bool | None = Field(
        default=None, description="Whether to filter by investor owned"
    )
    is_owner_occupied: bool | None = Field(
        default=None, description="Whether to filter by owner occupied"
    )

    # Market flags
    current_on_market_flag: bool | None = Field(
        default=None, description="Whether to filter by current on market flag"
    )
    current_on_market_rental_flag: bool | None = Field(
        default=None, description="Whether to filter by current on market rental flag"
    )
    current_new_construction_flag: bool | None = Field(
        default=None, description="Whether to filter by current new construction flag"
    )
    current_owner_occupied_flag: bool | None = Field(
        default=None, description="Whether to filter by current owner occupied flag"
    )
    current_investor_owned_flag: bool | None = Field(
        default=None, description="Whether to filter by current investor owned flag"
    )
    current_entity_owner_name: str | None = Field(
        default=None, description="Current entity owner name to filter by"
    )

    # Pagination
    limit: int | None = Field(
        default=None,
        ge=1,
        le=RequestLimits.PROPERTY_V2_MAX.value,
        description=f"Number of results to return (max: {RequestLimits.PROPERTY_V2_MAX.value})",
    )

    # Additional parameters
    params: dict[str, Any] | None = Field(
        default_factory=dict, description="Additional parameters to pass to the request"
    )

    @field_validator("property_types")
    @classmethod
    def validate_property_types(cls, v: list[str] | None) -> list[str] | None:
        """Validate property types against allowed values."""
        if v is not None:
            valid_types = [pt.value for pt in PropertyTypes]
            for prop_type in v:
                if prop_type.upper() not in valid_types:
                    raise ValueError(f"Invalid property type: {prop_type}")
            return [pt.upper() for pt in v]
        return v

    @field_validator("event_names")
    @classmethod
    def validate_event_names(cls, v: list[str] | None) -> list[str] | None:
        """Validate event names and convert to uppercase."""
        if v is not None:
            return [name.upper() for name in v]
        return v

    @field_validator("owner_name")
    @classmethod
    def validate_owner_names(cls, v: list[str] | None) -> list[str] | None:
        """Validate owner names and convert to uppercase."""
        if v is not None:
            return [name.upper() for name in v]
        return v

    @field_validator("entity_seller_name")
    @classmethod
    def validate_entity_seller_names(cls, v: list[str] | None) -> list[str] | None:
        """Validate entity seller names and convert to uppercase."""
        if v is not None:
            return [name.upper() for name in v]
        return v

    @field_validator(
        "min_record_added_date",
        "max_record_added_date",
        "min_event_date",
        "max_event_date",
        "min_record_updated_date",
        "max_record_updated_date",
    )
    @classmethod
    def validate_date_format(cls, v: str | None) -> str | None:
        """Validate date format is YYYY-MM-DD."""
        if v is not None:
            try:
                datetime.fromisoformat(v)
            except ValueError as err:
                raise ValueError(f"Date must be in YYYY-MM-DD format, got: {v}") from err
            else:
                return v
        return v

    @field_validator("min_beds", "max_beds")
    @classmethod
    def validate_bedroom_range(cls, v: int | None, info: ValidationInfo) -> int | None:
        """Validate bedroom range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_beds" and info.data.get("min_beds"):
                if v < info.data["min_beds"]:
                    raise ValueError("max_beds cannot be less than min_beds")
            elif field_name == "min_beds" and info.data.get("max_beds"):
                if v > info.data["max_beds"]:
                    raise ValueError("min_beds cannot be greater than max_beds")
        return v

    @field_validator("min_baths", "max_baths")
    @classmethod
    def validate_bathroom_range(cls, v: float | None, info: ValidationInfo) -> float | None:
        """Validate bathroom range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_baths" and info.data.get("min_baths"):
                if v < info.data["min_baths"]:
                    raise ValueError("max_baths cannot be less than min_baths")
            elif field_name == "min_baths" and info.data.get("max_baths"):
                if v > info.data["max_baths"]:
                    raise ValueError("min_baths cannot be greater than max_baths")
        return v

    @field_validator("min_sqft", "max_sqft")
    @classmethod
    def validate_sqft_range(cls, v: int | None, info: ValidationInfo) -> int | None:
        """Validate square footage range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_sqft" and info.data.get("min_sqft"):
                if v < info.data["min_sqft"]:
                    raise ValueError("max_sqft cannot be less than min_sqft")
            elif field_name == "min_sqft" and info.data.get("max_sqft"):
                if v > info.data["max_sqft"]:
                    raise ValueError("min_sqft cannot be greater than max_sqft")
        return v

    @field_validator("min_year_built", "max_year_built")
    @classmethod
    def validate_year_built_range(cls, v: int | None, info: ValidationInfo) -> int | None:
        """Validate year built range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_year_built" and info.data.get("min_year_built"):
                if v < info.data["min_year_built"]:
                    raise ValueError("max_year_built cannot be less than min_year_built")
            elif field_name == "min_year_built" and info.data.get("max_year_built"):
                if v > info.data["max_year_built"]:
                    raise ValueError("min_year_built cannot be greater than max_year_built")
        return v

    @field_validator("min_price", "max_price")
    @classmethod
    def validate_price_range(cls, v: int | None, info: ValidationInfo) -> int | None:
        """Validate price range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_price" and info.data.get("min_price"):
                if v < info.data["min_price"]:
                    raise ValueError("max_price cannot be less than min_price")
            elif field_name == "min_price" and info.data.get("max_price"):
                if v > info.data["max_price"]:
                    raise ValueError("min_price cannot be greater than max_price")
        return v

    @field_validator("min_record_added_date", "max_record_added_date")
    @classmethod
    def validate_record_added_date_range(cls, v: str | None, info: ValidationInfo) -> str | None:
        """Validate record added date range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_record_added_date" and info.data.get("min_record_added_date"):
                if v < info.data["min_record_added_date"]:
                    raise ValueError("max_record_added_date cannot be before min_record_added_date")
            elif field_name == "min_record_added_date" and info.data.get("max_record_added_date"):
                if v > info.data["max_record_added_date"]:
                    raise ValueError("min_record_added_date cannot be after max_record_added_date")
        return v

    @field_validator("min_event_date", "max_event_date")
    @classmethod
    def validate_event_date_range(cls, v: str | None, info: ValidationInfo) -> str | None:
        """Validate event date range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_event_date" and info.data.get("min_event_date"):
                if v < info.data["min_event_date"]:
                    raise ValueError("max_event_date cannot be before min_event_date")
            elif field_name == "min_event_date" and info.data.get("max_event_date"):
                if v > info.data["max_event_date"]:
                    raise ValueError("min_event_date cannot be after max_event_date")
        return v

    @field_validator("min_record_updated_date", "max_record_updated_date")
    @classmethod
    def validate_record_updated_date_range(cls, v: str | None, info: ValidationInfo) -> str | None:
        """Validate record updated date range consistency."""
        if v is not None and info.data:
            field_name = info.field_name
            if field_name == "max_record_updated_date" and info.data.get("min_record_updated_date"):
                if v < info.data["min_record_updated_date"]:
                    raise ValueError(
                        "max_record_updated_date cannot be before min_record_updated_date"
                    )
            elif field_name == "min_record_updated_date" and info.data.get(
                "max_record_updated_date"
            ):
                if v > info.data["max_record_updated_date"]:
                    raise ValueError(
                        "min_record_updated_date cannot be after max_record_updated_date"
                    )
        return v

    class Config:
        """Pydantic configuration."""

        extra = "forbid"  # Reject any extra fields
        validate_assignment = True  # Validate on assignment
        str_strip_whitespace = True  # Strip whitespace from strings


class PropertyV2RetrieveParamCategories(BaseModel):
    """High level categories for PropertyV2RetrieveParams."""

    property_filters: dict[str, Any] = Field(default_factory=dict, description="Property filters")
    event_filters: dict[str, Any] = Field(default_factory=dict, description="Event filters")
    owner_filters: dict[str, Any] = Field(default_factory=dict, description="Owner filters")



from datetime import UTC, datetime

from parcllabs.common import (
    ZIP_CODE_LENGTH,
)


class Validators:
    @staticmethod
    def validate_date(date_str: str) -> str | None:
        """
        Validates the date string and returns it in the 'YYYY-MM-DD' format.
        Raises ValueError if the date is invalid or not in the expected format.
        """
        if date_str:
            try:
                formatted_date = (
                    datetime.strptime(date_str, "%Y-%m-%d").replace(tzinfo=UTC).strftime("%Y-%m-%d")
                )
            except ValueError as e:
                raise ValueError(f"Date {date_str} is not in the correct format YYYY-MM-DD.") from e
            else:
                return formatted_date
        return None

    @staticmethod
    def validate_input_str_param(
        param: str,
        param_name: str,
        valid_values: list[str],
        params_dict: dict | None = None,
    ) -> dict | None:
        if param:
            param = param.upper()
            params_dict[param_name] = param

            if param not in valid_values:
                raise ValueError(
                    f"{param_name} value error. Valid values are: {valid_values}. Received: {param}"
                )

        return params_dict

    @staticmethod
    def validate_input_bool_param(
        param: object,
        param_name: str,
        params_dict: dict | None = None,
    ) -> dict | None:
        if param is None:
            return params_dict

        if not isinstance(param, bool):
            raise TypeError(f"{param_name} value error. Expected boolean. Received: {param}")

        params_dict[param_name] = "true" if param else "false"
        return params_dict

    @staticmethod
    def validate_input_bool_param_simple(param: bool | None) -> str:
        if not isinstance(param, bool):
            raise TypeError(f"Expected boolean. Received: {param}")

        return "true" if param else "false"

    @staticmethod
    def validate_us_zip_code(zip_code: str) -> str:
        """
        Validates the US zip code string and returns it in the expected format.
        Raises ValueError if the zip code is invalid or not in the expected format.
        """
        zip_code = zip_code.strip()
        if not zip_code.isdigit() or len(zip_code) != ZIP_CODE_LENGTH:
            raise ValueError(f"Zip code {zip_code} is not a valid 5-digit US zip code.")
        return zip_code

    @staticmethod
    def validate_field_exists(data: dict, fields: list[str]) -> None:
        """
        Validates that the required fields exist in the provided dictionary.
        """
        missing_fields = [field for field in fields if field not in data.keys()]
        if missing_fields:
            raise ValueError(
                f"Missing required fields: {', '.join(missing_fields)}. Provided request: {data}"
            )



from collections import deque

import pandas as pd

from parcllabs.exceptions import NotFoundError
from parcllabs.services.parcllabs_service import ParclLabsService
from parcllabs.services.validators import Validators


class PropertySearch(ParclLabsService):
    """
    Retrieve parcl_property_id for geographic markets in the Parcl Labs API.
    """

    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)

    def _handle_numeric_params(self, params: dict, **kwargs: object) -> dict:
        """Handle numeric parameters for property search."""
        numeric_params = [
            "bedrooms_max",
            "bedrooms_min",
            "bathrooms_max",
            "bathrooms_min",
            "year_built_max",
            "year_built_min",
            "square_footage_max",
            "square_footage_min",
        ]

        for param in numeric_params:
            if kwargs.get(param) is not None:
                params[param] = kwargs[param]

        return params

    def _handle_date_params(
        self,
        params: dict,
        record_added_date_start: str | None = None,
        record_added_date_end: str | None = None,
    ) -> dict:
        """Handle date parameters for property search."""
        if record_added_date_start:
            record_added_date_start = Validators.validate_date(record_added_date_start)
            params["record_added_date_start"] = record_added_date_start
        if record_added_date_end:
            record_added_date_end = Validators.validate_date(record_added_date_end)
            params["record_added_date_end"] = record_added_date_end

        return params

    def _prepare_params(
        self,
        property_type: str,
        square_footage_min: int | None = None,
        square_footage_max: int | None = None,
        bedrooms_min: int | None = None,
        bedrooms_max: int | None = None,
        bathrooms_min: int | None = None,
        bathrooms_max: int | None = None,
        year_built_min: int | None = None,
        year_built_max: int | None = None,
        current_entity_owner_name: str | None = None,
        event_history_sale_flag: bool | None = None,
        event_history_rental_flag: bool | None = None,
        event_history_listing_flag: bool | None = None,
        current_new_construction_flag: bool | None = None,
        current_owner_occupied_flag: bool | None = None,
        current_investor_owned_flag: bool | None = None,
        record_added_date_start: str | None = None,
        record_added_date_end: str | None = None,
        current_on_market_flag: bool | None = None,
    ) -> dict:
        params = {}

        if property_type:
            params["property_type"] = property_type.upper()

        if current_entity_owner_name:
            params["current_entity_owner_name"] = current_entity_owner_name.upper()

        # Handle boolean flags
        bool_flags = {
            "event_history_sale_flag": event_history_sale_flag,
            "event_history_rental_flag": event_history_rental_flag,
            "event_history_listing_flag": event_history_listing_flag,
            "current_new_construction_flag": current_new_construction_flag,
            "current_owner_occupied_flag": current_owner_occupied_flag,
            "current_investor_owned_flag": current_investor_owned_flag,
            "current_on_market_flag": current_on_market_flag,
        }

        for param_name, param_value in bool_flags.items():
            params = Validators.validate_input_bool_param(
                param=param_value,
                param_name=param_name,
                params_dict=params,
            )

        # Handle numeric parameters
        params = self._handle_numeric_params(
            params,
            bedrooms_max=bedrooms_max,
            bedrooms_min=bedrooms_min,
            bathrooms_max=bathrooms_max,
            bathrooms_min=bathrooms_min,
            year_built_max=year_built_max,
            year_built_min=year_built_min,
            square_footage_max=square_footage_max,
            square_footage_min=square_footage_min,
        )

        # Handle date parameters
        return self._handle_date_params(
            params,
            record_added_date_start=record_added_date_start,
            record_added_date_end=record_added_date_end,
        )

    def retrieve(
        self,
        parcl_ids: list[int],
        property_type: str,
        square_footage_min: int | None = None,
        square_footage_max: int | None = None,
        bedrooms_min: int | None = None,
        bedrooms_max: int | None = None,
        bathrooms_min: int | None = None,
        bathrooms_max: int | None = None,
        year_built_min: int | None = None,
        year_built_max: int | None = None,
        current_entity_owner_name: str | None = None,
        event_history_sale_flag: bool | None = None,
        event_history_rental_flag: bool | None = None,
        event_history_listing_flag: bool | None = None,
        current_new_construction_flag: bool | None = None,
        current_owner_occupied_flag: bool | None = None,
        current_investor_owned_flag: bool | None = None,
        record_added_date_start: str | None = None,
        record_added_date_end: str | None = None,
        current_on_market_flag: bool | None = None,
    ) -> pd.DataFrame:
        """
        Retrieve parcl_property_id for geographic markets based on specified criteria.

        Args:
            parcl_ids (List[int]): List of Parcl IDs for the geographic markets.
            property_type (str): The type of property to search for
            (e.g., 'single_family').
            square_footage_min (int, optional): Minimum square footage.
            Defaults to None.
            square_footage_max (int, optional): Maximum square footage.
            Defaults to None.
            bedrooms_min (int, optional): Minimum number of bedrooms. Defaults to None.
            bedrooms_max (int, optional): Maximum number of bedrooms. Defaults to None.
            bathrooms_min (int, optional): Minimum number of bathrooms.
            Defaults to None.
            bathrooms_max (int, optional): Maximum number of bathrooms.
            Defaults to None.
            year_built_min (int, optional): Minimum year built. Defaults to None.
            year_built_max (int, optional): Maximum year built. Defaults to None.
            current_entity_owner_name (str, optional): Filter by current owner entity
            name. Defaults to None.
            event_history_sale_flag (bool, optional): Filter properties with sale events
            in history. Defaults to None.
            event_history_rental_flag (bool, optional): Filter properties with rental
            events in history. Defaults to None.
            event_history_listing_flag (bool, optional): Filter properties with listing
            events in history. Defaults to None.
            current_new_construction_flag (bool, optional): Filter properties that are
            new construction. Defaults to None.
            current_owner_occupied_flag (bool, optional): Filter properties that are
            owner-occupied. Defaults to None.
            current_investor_owned_flag (bool, optional): Filter properties that are
            investor-owned. Defaults to None.
            record_added_date_start (str, optional): Filter properties added on or after
            this date (YYYY-MM-DD). Defaults to None.
            record_added_date_end (str, optional): Filter properties added on or before
            this date (YYYY-MM-DD). Defaults to None.
            current_on_market_flag (bool, optional): Filter properties currently on the
            market. Defaults to None.

        Returns:
            pd.DataFrame: A DataFrame containing the parcl_property_id and other detail
            for matching properties. Returns an empty DataFrame if no data is found for
            any market.
        """
        params = self._prepare_params(
            property_type=property_type,
            square_footage_min=square_footage_min,
            square_footage_max=square_footage_max,
            bedrooms_min=bedrooms_min,
            bedrooms_max=bedrooms_max,
            bathrooms_min=bathrooms_min,
            bathrooms_max=bathrooms_max,
            year_built_min=year_built_min,
            year_built_max=year_built_max,
            current_entity_owner_name=current_entity_owner_name,
            event_history_sale_flag=event_history_sale_flag,
            event_history_rental_flag=event_history_rental_flag,
            event_history_listing_flag=event_history_listing_flag,
            current_new_construction_flag=current_new_construction_flag,
            current_owner_occupied_flag=current_owner_occupied_flag,
            current_investor_owned_flag=current_investor_owned_flag,
            record_added_date_start=record_added_date_start,
            record_added_date_end=record_added_date_end,
            current_on_market_flag=current_on_market_flag,
        )

        output_data = deque()
        markets_with_no_data = []

        for parcl_id in parcl_ids:
            try:
                params["parcl_id"] = parcl_id
                response = self._get(url=self.full_url, params=params)
                data = response.json()
                df_container = pd.DataFrame(data.get("items"))
                self._update_account_info(data.get("account"))
                output_data.append(df_container)
            except NotFoundError:
                # Track markets with no data
                markets_with_no_data.append(parcl_id)
                continue
            except Exception:
                # For other exceptions, simply re-raise without specifying
                raise

        if not output_data:
            # If no data was found for any market, return empty DataFrame
            return pd.DataFrame()

        results = pd.concat(output_data).reset_index(drop=True)

        # Print message about markets with no data if any exist
        if markets_with_no_data:
            print(
                f"No data found for markets with parcl_ids: "
                f"{', '.join(map(str, markets_with_no_data))}"
            )

        return results



from parcllabs.__version__ import VERSION as __version__  # noqa: F401, N811

# Constants
DEFAULT_API_BASE = "https://api.parcllabs.com"

api_key: str | None = None
api_base = DEFAULT_API_BASE

from parcllabs.parcllabs_client import ParclLabsClient  # noqa: E402, F401
